/* fft.h */

typedef float fft_type;

/* Initialise fft with the size of the data */
void fft_init(int);
void fft_data_float(float *);
void fft_data_signed16(int16_t *);
void fft_data_unsigned16(uint16_t *);
fft_type *fft_getresult(void);
void fft_window(void);
void fft_compute(void);
