/* map3d.h */

void map3d_init(void);

void map3d_close(void);
int map3d_load(void);
void map3d_save(void);
void map3d_transform(int, double, double, double, int *, int *);
void map3d_setcal(int, int, double, double, double, int, int);
int map3d_calibrate(int);
void map3d_setparams(int, int, int, double, double, double);
