/* abi.h */

typedef int (*vm_intfn)(void);

/*
 * We must include here all prototypes for functions which can be
 * included in the ABI function table
 */

int vm_intfn_nop(void);
int vm_intfn_global_store(void);
int vm_intfn_global_load(void);
int vm_intfn_global_array_store(void);
int vm_intfn_global_array_load(void);
int vm_intfn_printint(void);
int vm_intfn_printreal(void);
int vm_intfn_printstr(void);
int vm_intfn_dmxsetchannel(void);
int vm_intfn_dmxoutput(void);
int vm_intfn_gettime(void);
int vm_intfn_waittime(void);
int vm_intfn_wait(void);
int vm_intfn_wakeup(void);
int vm_intfn_spawn(void);
int vm_intfn_midi_read(void);
int vm_intfn_beatdetect_read(void);
int vm_intfn_beatdetect_phase(void);
int vm_intfn_beatdetect_confidence(void);
int vm_intfn_realtoint(void);
int vm_intfn_inttoreal(void);
int vm_intfn_map3d_setcal(void);
int vm_intfn_map3d_calibrate(void);
int vm_intfn_map3d_transform(void);
int vm_intfn_map3d_setparams(void);
int vm_intfn_map3d_load(void);
int vm_intfn_map3d_save(void);
int vm_intfn_sin(void);
int vm_intfn_cos(void);
int vm_intfn_random(void);
int vm_intfn_mouse_read(void);
