/* plugins.h */

/*
 * The master table of plugins is found in plugins.c. This is
 * just a declaration to allow the table to be used.
 */

#include "plugin.h"

extern struct plugin plugins_table[];

extern int nplugins;

